# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
import warnings
from abc import ABC
from datetime import datetime

from . import _date_to_ISO, _ISO_to_date


class _TaipyBase(ABC):
    def __init__(self, data: t.Any, hash_name: str) -> None:
        self.__data = data
        self.__hash_name = hash_name

    def get(self):
        return self.__data

    def get_name(self):
        return self.__hash_name

    def set(self, data: t.Any):
        self.__data = data

    def cast_value(self, value: t.Any):
        return value

    @staticmethod
    def get_hash():
        return "TaipyBase"


class _TaipyData(_TaipyBase):
    @staticmethod
    def get_hash():
        return "_TpD"


class _TaipyBool(_TaipyBase):
    def get(self):
        return self.cast_value(super().get())

    def cast_value(self, value: t.Any):
        return bool(value)

    @staticmethod
    def get_hash():
        return "_TpB"


class _TaipyNumber(_TaipyBase):
    def get(self):
        try:
            return float(super().get())
        except ValueError:
            raise ValueError(f"Variable {self.get_name()} should hold a number")

    def cast_value(self, value: t.Any):
        if isinstance(value, str):
            try:
                return float(value) if value else 0.0
            except Exception as e:
                warnings.warn(f"{self.get_name()}: Parsing {value} as float:\n{e}")
                return 0.0
        return super().cast_value(value)

    @staticmethod
    def get_hash():
        return "_TpN"


class _TaipyDate(_TaipyBase):
    def get(self):
        val = super().get()
        if isinstance(val, datetime):
            val = _date_to_ISO(val)
        elif val is not None:
            val = str(val)
        return val

    def cast_value(self, value: t.Any):
        if isinstance(value, str):
            return _ISO_to_date(value)
        return super().cast_value(value)

    @staticmethod
    def get_hash():
        return "_TpDt"


class _TaipyLovValue(_TaipyBase):
    @staticmethod
    def get_hash():
        return "_TpLv"


class _TaipyLov(_TaipyBase):
    @staticmethod
    def get_hash():
        return "_TpL"


class _TaipyContent(_TaipyBase):
    @staticmethod
    def get_hash():
        return "_TpC"


class _TaipyContentImage(_TaipyBase):
    @staticmethod
    def get_hash():
        return "_TpCi"



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
from datetime import datetime
from random import random

from ..data.data_scope import _DataScopes
from ._map_dict import _MapDict

if t.TYPE_CHECKING:
    from ..gui import Gui


class _Bindings:
    def __init__(self, gui: "Gui") -> None:
        self.__gui = gui
        self.__scopes = _DataScopes()

    def _bind(self, name: str, value: t.Any) -> None:
        if hasattr(self, name):
            raise ValueError(f"Variable '{name}' is already bound")
        if not name.isidentifier():
            raise ValueError(f"Variable name '{name}' is invalid")
        if isinstance(value, dict):
            setattr(self._get_data_scope(), name, _MapDict(value))
        else:
            setattr(self._get_data_scope(), name, value)
        # prop = property(self.__value_getter(name), self.__value_setter(name))
        setattr(_Bindings, name, self.__get_property(name))

    def __get_property(self, name):
        def __setter(ud: _Bindings, value: t.Any):
            if isinstance(value, dict):
                value = _MapDict(value, None)
            ud.__gui._update_var(name, value)

        def __getter(ud: _Bindings) -> t.Any:
            value = getattr(ud._get_data_scope(), name)
            if isinstance(value, _MapDict):
                return _MapDict(value._dict, lambda k, v: ud.__gui._update_var(f"{name}.{k}", v))
            else:
                return value

        return property(__getter, __setter)  # Getter, Setter

    def _set_single_client(self, value: bool) -> None:
        self.__scopes.set_single_client(value)

    def _get_single_client(self) -> bool:
        return self.__scopes.get_single_client()

    def _set_client_id(self, message: dict):
        self.__scopes._set_client_id(message.get("client_id"))

    def _get_or_create_scope(self, id: str):
        if not id:
            id = f"{datetime.now().strftime('%Y%m%d%H%M%S%f')}-{random()}"
            self.__gui._send_ws_id(id)
        self.__scopes.create_scope(id)

    def _new_scopes(self):
        self.__scopes = _DataScopes()

    def _get_data_scope(self):
        return self.__scopes.get_scope()

    def _get_all_scopes(self):
        return self.__scopes.get_all_scopes()



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from markdown.inlinepatterns import InlineProcessor

from .factory import _MarkdownFactory


class _ControlPattern(InlineProcessor):

    __PATTERN = _MarkdownFactory._TAIPY_START + "([a-zA-Z][\\.a-zA-Z_$0-9]*)(.*?)" + _MarkdownFactory._TAIPY_END

    @staticmethod
    def extend(md, gui, priority):
        instance = _ControlPattern(_ControlPattern.__PATTERN, md)
        md.inlinePatterns.register(instance, "taipy", priority)
        instance._gui = gui

    def handleMatch(self, m, data):
        return _MarkdownFactory.create_element(self._gui, m.group(1), m.group(2)), m.start(0), m.end(0)



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import inspect
import typing as t
import warnings
from abc import ABC, abstractmethod

from ..utils import _TaipyData
from .data_format import _DataFormat


class _DataAccessor(ABC):

    _WS_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

    @staticmethod
    @abstractmethod
    def get_supported_classes() -> t.List[str]:
        pass

    @abstractmethod
    def get_data(
        self, guiApp: t.Any, var_name: str, value: t.Any, payload: t.Dict[str, t.Any], data_format: _DataFormat
    ) -> t.Dict[str, t.Any]:
        pass

    @abstractmethod
    def get_col_types(self, var_name: str, value: t.Any) -> t.Dict[str, str]:
        pass


class _InvalidDataAccessor(_DataAccessor):
    @staticmethod
    def get_supported_classes() -> t.List[str]:
        return [type(None).__name__]

    def get_data(
        self, guiApp: t.Any, var_name: str, value: t.Any, payload: t.Dict[str, t.Any], data_format: _DataFormat
    ) -> t.Dict[str, t.Any]:
        return {}

    def get_col_types(self, var_name: str, value: t.Any) -> t.Dict[str, str]:
        return {}


class _DataAccessors(object):
    def __init__(self) -> None:
        self.__access_4_type: t.Dict[str, _DataAccessor] = {}

        self.__invalid_data_accessor = _InvalidDataAccessor()

        self.__data_format = _DataFormat.JSON

        from .array_dict_data_accessor import _ArrayDictDataAccessor
        from .numpy_data_accessor import _NumpyDataAccessor
        from .pandas_data_accessor import _PandasDataAccessor

        self._register(_PandasDataAccessor)
        self._register(_ArrayDictDataAccessor)
        self._register(_NumpyDataAccessor)

    def _register(self, cls: t.Type[_DataAccessor]) -> None:
        if inspect.isclass(cls):
            if issubclass(cls, _DataAccessor):
                names = cls.get_supported_classes()
                if not names:
                    raise TypeError(f"method {cls.__name__}.get_supported_classes returned an invalid value")
                # check existence
                inst: t.Optional[_DataAccessor] = None
                for name in names:
                    inst = self.__access_4_type.get(name)
                    if inst:
                        break
                if inst is None:
                    try:
                        inst = cls()
                    except Exception as e:
                        raise TypeError(f"Class {cls.__name__} cannot be instanciated") from e
                    if inst:
                        for name in names:
                            self.__access_4_type[name] = inst  # type: ignore
            else:
                raise TypeError(f"Class {cls.__name__} is not a subclass of DataAccessor")
        else:
            raise AttributeError("The argument of 'DataAccessors.register' should be a class")

    def __get_instance(self, value: _TaipyData) -> _DataAccessor:  # type: ignore
        value = value.get()
        access = self.__access_4_type.get(type(value).__name__)
        if access is None:
            warnings.warn(f"Can't find Data Accessor for type {type(value).__name__}")
            return self.__invalid_data_accessor
        return access

    def _get_data(
        self, guiApp: t.Any, var_name: str, value: _TaipyData, payload: t.Dict[str, t.Any]
    ) -> t.Dict[str, t.Any]:
        return self.__get_instance(value).get_data(guiApp, var_name, value.get(), payload, self.__data_format)

    def _get_col_types(self, var_name: str, value: _TaipyData) -> t.Dict[str, str]:
        return self.__get_instance(value).get_col_types(var_name, value.get())

    def _set_data_format(self, data_format: _DataFormat):
        self.__data_format = data_format



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from ..factory import _Factory


class _MarkdownFactory(_Factory):
    # Taipy Markdown tags
    _TAIPY_START = "TaIpY:"
    _TAIPY_END = ":tAiPy"

    _TAIPY_BLOCK_TAGS = ["layout", "part", "expandable", "dialog", "pane"]

    @staticmethod
    def create_element(gui, control_type: str, all_properties: str) -> str:
        # Create properties dict from all_properties
        property_pairs = _Factory._PROPERTY_RE.findall(all_properties)
        properties = {property[0]: property[1] for property in property_pairs}
        if builder := _Factory.CONTROL_BUILDERS[control_type](gui, control_type, properties):
            return builder.el
        else:
            return f"<|INVALID SYNTAX - Control is '{control_type}'|>"



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t


class Icon:
    """Small image in the User Interface.

    Icons are typically used in controls like [button](../gui/viselements/button.md)
    or items in a [menu](../gui/viselements/menu.md).

    Attributes:
        path (str): The path to the image file.
        text (Optional[str]): The text associated to the image or None if there is none.

    If a text is associated to an icon, it is rendered by the visual elements that
    uses this Icon.
    """

    @staticmethod
    def get_dict_or(value: t.Union[str, t.Any]) -> t.Union[str, dict]:
        return value._to_dict() if isinstance(value, Icon) else value

    def __init__(self, path: str, text: t.Optional[str] = None) -> None:
        """Initialize a new Icon.

        Arguments:
            path (str): The path to an image file.
            text (Optional[str]): The text associated to the image. If _text_ is None,
                there is no text associated to this image.
        """
        self.path = path
        self.text = text

    def _to_dict(self, a_dict: t.Optional[dict] = None) -> dict:
        if a_dict is None:
            a_dict = {}
        a_dict["path"] = self.path
        if self.text is not None:
            a_dict["text"] = self.text
        return a_dict



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import sys
import threading


class _KillableThread(threading.Thread):  # pragma: no cover
    def __init__(self, *args, **keywords):
        threading.Thread.__init__(self, *args, **keywords)
        self.killed = False

    def start(self):
        self.__run_backup = self.run
        self.run = self.__run
        threading.Thread.start(self)

    def __run(self):
        sys.settrace(self.globaltrace)
        self.__run_backup()
        self.run = self.__run_backup

    def globaltrace(self, frame, why, arg):
        return self.localtrace if why == "call" else None

    def localtrace(self, frame, why, arg):
        if self.killed and why == "line":
            raise SystemExit(0)
        return self.localtrace

    def kill(self):
        self.killed = True



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import logging
import os
import re
import typing as t
import webbrowser

import __main__
from flask import Blueprint, Flask, json, jsonify, render_template, render_template_string, request, send_from_directory
from flask_cors import CORS
from flask_socketio import SocketIO
from flask_talisman import Talisman
from werkzeug.serving import is_running_from_reloader

from .renderers.jsonencoder import _TaipyJsonEncoder
from .utils import _is_in_notebook, _KillableThread

if t.TYPE_CHECKING:
    from .gui import Gui


class _Server:

    __RE_JSX_RENDER_ROUTE = re.compile(r"/taipy-jsx/(.*)/")
    __RE_OPENING_CURLY = re.compile(r"([^\"])(\{)")
    __RE_CLOSING_CURLY = re.compile(r"(\})([^\"])")
    __OPENING_CURLY = r"\1&#x7B;"
    __CLOSING_CURLY = r"&#x7D;\2"

    def __init__(
        self,
        gui: Gui,
        flask: t.Optional[Flask] = None,
        css_file: str = "",
        path_mapping: t.Optional[dict] = {},
        content_security_policy: t.Optional[dict] = None,
        force_https: bool = False,
    ):
        self._gui = gui
        self._root_page_name = gui._get_root_page_name()
        self._flask = Flask("Taipy") if flask is None else flask
        self.css_file = css_file
        if "SECRET_KEY" not in self._flask.config or not self._flask.config["SECRET_KEY"]:
            self._flask.config["SECRET_KEY"] = "TaIpY"
        # set json encoder (for Taipy specific types)
        self._flask.json_encoder = _TaipyJsonEncoder
        # Add cors for frontend access
        self._ws = SocketIO(
            self._flask,
            async_mode=None,
            cors_allowed_origins="*",
            ping_timeout=10,
            ping_interval=5,
            json=json,
        )
        # this is necessary since CORS resources can't been None eventhough python stub allows for None
        CORS(self._flask)

        if force_https or content_security_policy:
            Talisman(self._flask, content_security_policy=content_security_policy, force_https=force_https)

        self.__path_mapping = path_mapping

        # Websocket (handle json message)
        @self._ws.on("message")
        def handle_message(message) -> None:
            if "status" in message:
                print(message["status"])
            elif "type" in message.keys():
                gui._manage_message(message["type"], message)

    def __get_client_config(self) -> t.Dict[str, t.Any]:
        config = {
            "timeZone": self._gui._config.get_time_zone(),
            "darkMode": self._gui._get_config("dark_mode", True),
            }
        themes=self._gui._get_themes()
        if themes:
            config["themes"] = themes
        return config
        

    def _get_default_blueprint(
        self,
        static_folder: t.Optional[str] = "",
        template_folder: str = "",
        title: str = "",
        favicon: str = "",
        root_margin: t.Optional[str] = None,
    ) -> Blueprint:
        taipy_bp = Blueprint("Taipy", __name__, static_folder=static_folder, template_folder=template_folder)
        # Serve static react build

        @taipy_bp.route("/", defaults={"path": ""})
        @taipy_bp.route("/<path:path>")
        def my_index(path):
            if path == "" or "." not in path:
                return render_template(
                    "index.html",
                    app_css=f"/{self.css_file}.css",
                    title=title,
                    favicon=favicon,
                    root_margin=root_margin,
                    watermark=self._gui._get_config("watermark", None),
                    config=self.__get_client_config()

                )
            if os.path.isfile(static_folder + os.path.sep + path):
                return send_from_directory(static_folder + os.path.sep, path)
            # use the path mapping to detect and find resources
            for k, v in self.__path_mapping.items():
                if path.startswith(f"{k}/") and os.path.isfile(v + os.path.sep + path[len(k) + 1 :]):
                    return send_from_directory(v + os.path.sep, path[len(k) + 1 :])
            if hasattr(__main__, "__file__") and os.path.isfile(
                os.path.dirname(__main__.__file__) + os.path.sep + path
            ):
                return send_from_directory(os.path.dirname(__main__.__file__) + os.path.sep, path)
            if os.path.isfile(self._gui._root_dir + os.path.sep + path):
                return send_from_directory(self._gui._root_dir + os.path.sep, path)
            return ("", 404)

        @taipy_bp.errorhandler(404)
        def page_not_found(e):
            return f"{e.message}, {e.description}"

        return taipy_bp

    # Update to render as JSX
    def _render(self, html_fragment, style, head):
        template_str = _Server.__RE_OPENING_CURLY.sub(_Server.__OPENING_CURLY, html_fragment)
        template_str = _Server.__RE_CLOSING_CURLY.sub(_Server.__CLOSING_CURLY, template_str)
        template_str = template_str.replace('"{!', "{")
        template_str = template_str.replace('!}"', "}")
        return self._direct_render_json(
            {
                "jsx": template_str,
                "style": (style + os.linesep) if style else "",
                "head": head or [],
            }
        )

    def _direct_render_json(self, data):
        return jsonify(data)

    def _render_page(self) -> t.Any:
        page = None
        render_path_name = _Server.__RE_JSX_RENDER_ROUTE.match(request.path).group(1)  # type: ignore
        # Get page instance
        for page_i in self._gui._config.pages:
            if page_i._route == render_path_name:
                page = page_i
        # try partials
        if page is None:
            page = self._gui._get_partial(render_path_name)
        # Make sure that there is a page instance found
        if page is None:
            return (jsonify({"error": "Page doesn't exist!"}), 400, {"Content-Type": "application/json; charset=utf-8"})
        page.render(self._gui)
        if (
            render_path_name == self._root_page_name
            and page._rendered_jsx is not None
            and "<PageContent" not in page._rendered_jsx
        ):
            page._rendered_jsx += "<PageContent />"
        # Return jsx page
        if page._rendered_jsx is not None:
            return self._render(page._rendered_jsx, page._style if page._style is not None else "", page._head)
        else:
            return ("No page template", 404)

    def _render_route(self) -> t.Any:
        router = '<Routes key="routes">'
        router += (
            '<Route path="/" key="'
            + self._root_page_name
            + '" element={<MainPage key="tr'
            + self._root_page_name
            + '" path="/'
            + self._root_page_name
            + '"'
        )
        routes = self._gui._config.routes
        route = next((r for r in routes if r != self._root_page_name), None)
        router += (' route="/' + route + '"') if route else ""
        router += " />} >"
        locations = {"/": f"/{self._root_page_name}"}
        for route in routes:
            if route != self._root_page_name:
                router += (
                    '<Route path="'
                    + route
                    + '" key="'
                    + route
                    + '" element={<TaipyRendered key="tr'
                    + route
                    + '"/>} />'
                )
                locations[f"/{route}"] = f"/{route}"
        router += '<Route path="*" key="NotFound" element={<NotFound404 />} />'
        router += "</Route>"
        router += "</Routes>"

        return self._direct_render_json(
            {
                "router": router,
                "locations": locations,
                "timeZone": self._gui._config.get_time_zone(),
                "darkMode": self._gui._get_config("dark_mode", True),
                "blockUI": self._gui._is_ui_blocked(),
            }
        )

    def get_flask(self):
        return self._flask

    def test_client(self):
        return self._flask.test_client()

    def _run_notebook(self):
        self._ws.run(self._flask, host=self._host, port=self._port, debug=False, use_reloader=False)

    def runWithWS(self, host, port, debug, use_reloader, flask_log, run_in_thread):
        host_value = host if host != "0.0.0.0" else "localhost"
        if not flask_log:
            log = logging.getLogger("werkzeug")
            log.disabled = True
            self._flask.logger.disabled = True
            print(f" * Server starting on http://{host_value}:{port}")
        if (
            not is_running_from_reloader()
            and self._gui._get_config("run_browser", False)
            and self._gui._get_config("debug", False)
        ):
            webbrowser.open(f"http://{host_value}{f':{port}' if port else ''}", new=2)
        if _is_in_notebook() or run_in_thread:
            self._host = host
            self._port = port
            self._thread = _KillableThread(target=self._run_notebook)
            self._thread.start()
            return
        self._ws.run(self._flask, host=host, port=port, debug=debug, use_reloader=use_reloader)



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from .config import Config

# Default config loaded by app.py
default_config: Config = {
    "port": 5000,
    "dark_mode": True,
    "debug": True,
    "host": "127.0.0.1",
    "use_reloader": True,
    "time_zone": None,
    "propagate": True,
    "favicon": None,
    "title": None,
    "theme": None,
    "light_theme": None,
    "dark_theme": None,
    "use_arrow": False,
    "system_notification": False,
    "notification_duration": 3000,
    "single_client": False,
    "ngrok_token": "",
    "upload_folder": None,
    "flask_log": False,
    "margin": "1em",
    "run_browser": True,
    "content_security_policy": None,
    "force_https": False,
    "watermark": "Taipy inside",
}



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import contextlib
import json
import numbers
import re
import typing as t
import warnings
import xml.etree.ElementTree as etree
from datetime import date, datetime, time
from inspect import isclass

from ..partial import Partial
from ..types import _AttributeType, _get_taipy_type
from ..utils import (
    _date_to_ISO,
    _get_client_var_name,
    _get_data_type,
    _get_expr_var_name,
    _getscopeattr,
    _getscopeattr_drill,
    _is_boolean,
    _is_boolean_true,
    _MapDict,
)
from ..utils.types import _TaipyData
from .jsonencoder import _TaipyJsonEncoder
from .utils import (
    _add_to_dict_and_get,
    _get_col_from_indexed,
    _get_columns_dict,
    _get_idx_from_col,
    _get_tuple_val,
    _to_camel_case,
)


class _Builder:

    __keys: t.Dict[str, int] = {}

    __ONE_COLUMN_CHART = ["pie"]

    __BLOCK_CONTROLS = ["dialog", "expandable", "pane", "part"]

    def __init__(
        self,
        gui,
        control_type: str,
        element_name: str,
        attributes: t.Union[t.Dict[str, t.Any], None],
        default_value="<Empty>",
    ):
        from ..gui import Gui
        from .factory import _Factory

        self.el = etree.Element(element_name)

        self.__control_type = control_type
        self.__element_name = element_name
        self.__attributes = attributes or {}
        self.__hashes = {}
        self.__update_vars: t.List[str] = []
        self.__gui: Gui = gui

        self.__default_property_name = _Factory.get_default_property_name(control_type) or ""
        default_property_value = self.__attributes.get(self.__default_property_name, None)
        if default_property_value is None and default_value is not None:
            self.__attributes[self.__default_property_name] = default_value

        # Bind properties dictionary to attributes if condition is matched (will leave the binding for function at the builder )
        if "properties" in self.__attributes:
            (prop_dict, prop_hash) = self.__parse_attribute_value(self.__attributes["properties"])
            if prop_hash is None:
                prop_hash = prop_dict
                self.__gui._bind_var(prop_hash)
                if hasattr(self.__gui._bindings(), prop_hash):
                    prop_dict = _getscopeattr(self.__gui, prop_hash)
            if isinstance(prop_dict, _MapDict):
                # Iterate through prop_dict and append to self.attributes
                for k, v in prop_dict.items():
                    (val, key_hash) = self.__parse_attribute_value(v)
                    self.__attributes[k] = f"{{{prop_hash}['{k}']}}" if key_hash is None else v
            else:
                warnings.warn(f"{self.__control_type}.properties ({prop_hash}) must be a dict.")

        # Bind potential function and expressions in self.attributes
        for k, v in self.__attributes.items():
            val = v
            hashname = None
            if callable(v):
                if v.__name__ == "<lambda>":
                    hashname = _get_expr_var_name(v.__code__)
                    self.__gui._bind_var_val(hashname, v)
                else:
                    hashname = _get_expr_var_name(v.__name__)
            elif isinstance(v, str):
                # need to unescape the double quotes that were escaped during preprocessing
                (val, hashname) = self.__parse_attribute_value(v.replace('\\"', '"'))

            if val is not None or hashname:
                self.__attributes[k] = val
            if hashname:
                self.__hashes[k] = hashname
        # set classname
        self.__set_class_names()
        # define a unique key
        self.set_attribute("key", _Builder._get_key(self.__element_name))

    @staticmethod
    def __to_string(x: t.Any) -> str:
        return str(x)

    @staticmethod
    def _get_key(name: str) -> str:
        key_index = _Builder.__keys.get(name, 0)
        _Builder.__keys[name] = key_index + 1
        return f"{name}.{key_index}"

    @staticmethod
    def _reset_key() -> None:
        _Builder.__keys = {}

    def __get_list_of_(self, name: str):
        lof = self.__attributes.get(name)
        if isinstance(lof, str):
            lof = [s for s in lof.split(";")]
        return lof

    def __get_name_indexed_property(self, name: str) -> t.Dict[str, t.Any]:
        ret = {}
        index_re = re.compile(name + r"\[(.*)\]$")
        for key in self.__attributes.keys():
            if m := index_re.match(key):
                ret[m.group(1)] = self.__attributes.get(key)
        return ret

    def __get_multiple_indexed_attributes(
        self, names: t.Tuple[str], index: t.Optional[int] = None
    ) -> t.List[t.Optional[str]]:
        names = names if index is None else [f"{n}[{index}]" for n in names]  # type: ignore
        return [self.__attributes.get(name) for name in names]

    def __parse_attribute_value(self, value) -> t.Tuple:
        if isinstance(value, str) and self.__gui._is_expression(value):
            hash_value = self.__gui._evaluate_expr(value)
            try:
                func = self.__gui._get_user_function(hash_value)
                if callable(func):
                    return (func, hash_value)
                return (_getscopeattr_drill(self.__gui, hash_value), hash_value)
            except AttributeError:
                warnings.warn(f"Expression '{value}' cannot be evaluated")
        return (value, None)

    def __get_boolean_attribute(self, name: str, default_value=False):
        boolattr = self.__attributes.get(name, default_value)
        return _is_boolean_true(boolattr) if isinstance(boolattr, str) else bool(boolattr)

    def __set_boolean_attribute(self, name: str, value: bool):
        return self.__set_react_attribute(_to_camel_case(name), value)

    def __set_dict_attribute(self, name: str):
        dict_attr = self.__attributes.get(name)
        if dict_attr:
            if isinstance(dict_attr, str):
                vals = [x.strip().split(":") for x in dict_attr.split(";")]
                dict_attr = {}
                for val in vals:
                    if len(val) > 1:
                        dict_attr[val[0].strip()] = val[1].strip()
            if isinstance(dict_attr, (dict, _MapDict)):
                self.__set_json_attribute(_to_camel_case(name), dict_attr)
            else:
                warnings.warn(f"{self.__element_name} {name} should be a dict\n'{str(dict_attr)}'")
        return self

    def __set_json_attribute(self, name, value):
        return self.set_attribute(name, json.dumps(value, cls=_TaipyJsonEncoder))

    def __set_list_of_(self, name: str):
        lof = self.__get_list_of_(name)
        if not isinstance(lof, (list, tuple)):
            if lof is not None:
                warnings.warn(f"{self.__element_name} {name} should be a list")
            return self
        return self.__set_json_attribute(_to_camel_case(name), lof)

    def __set_number_attribute(
        self, name: str, default_value: t.Optional[str] = None, optional: t.Optional[bool] = True
    ):
        value = self.__attributes.get(name, default_value)
        if value is None:
            if not optional:
                warnings.warn(f"Property {name} is required for control {self.__control_type}")
            return self
        try:
            val = float(value)
        except ValueError:
            raise ValueError(f"Property {name} expects a number for control {self.__control_type}")
        return self.__set_react_attribute(_to_camel_case(name), val)

    def __set_string_attribute(
        self, name: str, default_value: t.Optional[str] = None, optional: t.Optional[bool] = True
    ):
        strattr = self.__attributes.get(name, default_value)
        if strattr is None:
            if not optional:
                warnings.warn(f"Property {name} is required for control {self.__control_type}")
            return self
        return self.set_attribute(_to_camel_case(name), str(strattr))

    def __set_dynamic_string_attribute(
        self, name: str, default_value: t.Optional[str] = None, with_update: t.Optional[bool] = False
    ):
        hash_name = self.__hashes.get(name)
        str_val = self.__attributes.get(name, default_value)
        if str_val is not None:
            self.set_attribute(_to_camel_case(f"default_{name}"), str(str_val))
        if hash_name:
            if with_update:
                self.__update_vars.append(f"{name}={hash_name}")
            self.__set_react_attribute(name, hash_name)
        return self

    def __set_function_attribute(
        self, name: str, default_value: t.Optional[str] = None, optional: t.Optional[bool] = True
    ):
        strattr = self.__attributes.get(name, default_value)
        if strattr is None:
            if not optional:
                warnings.warn(f"Property {name} is required for control {self.__control_type}")
            return self
        elif callable(strattr):
            strattr = self.__hashes.get(name)
            if strattr is None:
                return self
        elif strattr:
            strattr = str(strattr)
            func = self.__gui._get_user_function(strattr)
            if func == strattr:
                warnings.warn(f" {self.__control_type}.{name}: {strattr} is not a function")
        return self.set_attribute(_to_camel_case(name), strattr) if strattr else self

    def __set_string_or_number_attribute(self, name: str, default_value: t.Optional[t.Any] = None):
        attr = self.__attributes.get(name, default_value)
        if attr is None:
            return self
        if isinstance(attr, numbers.Number):
            return self.__set_react_attribute(_to_camel_case(name), attr)
        else:
            return self.set_attribute(_to_camel_case(name), attr)

    def __set_react_attribute(self, name: str, value: t.Any):
        return self.set_attribute(name, "{!" + (str(value).lower() if isinstance(value, bool) else str(value)) + "!}")

    def get_adapter(self, var_name: str, property_name: t.Optional[str] = None, multi_selection=True):  # noqa: C901
        property_name = var_name if property_name is None else property_name
        lov = self.__get_list_of_(var_name)
        if isinstance(lov, list):
            adapter = self.__attributes.get("adapter")
            if adapter and isinstance(adapter, str):
                adapter = self.__gui._get_user_function(adapter)
            if adapter and not callable(adapter):
                warnings.warn("'adapter' property value is invalid")
                adapter = None
            var_type = self.__attributes.get("type")
            if isclass(var_type):
                var_type = var_type.__name__  # type: ignore
            if not isinstance(var_type, str):
                elt = None
                if len(lov) == 0:
                    value = self.__attributes.get("value")
                    if isinstance(value, list):
                        if len(value) > 0:
                            elt = value[0]
                    else:
                        elt = value
                else:
                    elt = lov[0]
                var_type = self.__gui._get_unique_type_adapter(type(elt).__name__)
            if adapter is None:
                adapter = self.__gui._get_adapter_for_type(var_type)
            if lov_name := self.__hashes.get(var_name):
                if adapter is None:
                    adapter = self.__gui._get_adapter_for_type(lov_name)
                else:
                    self.__gui._add_type_for_var(lov_name, var_type)
            if value_name := self.__hashes.get("value"):
                if adapter is None:
                    adapter = self.__gui._get_adapter_for_type(value_name)
                else:
                    self.__gui._add_type_for_var(value_name, var_type)
            if adapter is not None:
                self.__gui._add_adapter_for_type(var_type, adapter)  # type: ignore

            ret_list = []
            if len(lov) > 0:
                for elt in lov:
                    ret = self.__gui._run_adapter(
                        t.cast(t.Callable, adapter), elt, adapter.__name__ if callable(adapter) else "adapter")  # type: ignore
                    if ret is not None:
                        ret_list.append(ret)
            self.__attributes[f"default_{property_name}"] = ret_list

            ret_list = []
            value = self.__attributes.get("value")
            val_list = value if isinstance(value, list) else [value]
            for val in val_list:
                ret = self.__gui._run_adapter(
                    t.cast(t.Callable, adapter), val, adapter.__name__ if callable(adapter) else "adapter", id_only=True)  # type: ignore
                if ret is not None:
                    ret_list.append(ret)
            if multi_selection:
                self.__set_default_value("value", ret_list)
            else:
                ret_val = ret_list[0] if len(ret_list) else ""
                if ret_val == "-1" and self.__attributes.get("unselected_value") is not None:
                    ret_val = str(self.__attributes.get("unselected_value", ""))
                self.__set_default_value("value", ret_val)
        return self

    def __update_col_desc_from_indexed(self, columns: t.Dict[str, t.Any], name: str):
        col_value = self.__get_name_indexed_property(name)
        for k, v in col_value.items():
            if col_desc := next((x for x in columns.values() if x["dfid"] == k), None):
                if col_desc.get(_to_camel_case(name)) is None:
                    col_desc[_to_camel_case(name)] = str(v)
            else:
                warnings.warn(f"{self.__element_name} {name}[{k}] is not in the list of displayed columns")

    def get_dataframe_attributes(self, date_format="MM/dd/yyyy", number_format=None):  # noqa: C901
        data = self.__attributes.get("data")
        data_hash = self.__hashes.get("data", "")
        col_types = self.__gui._accessors._get_col_types(data_hash, _TaipyData(data, data_hash))
        columns = _get_columns_dict(
            data,
            _add_to_dict_and_get(self.__attributes, "columns", {}),
            col_types,
            _add_to_dict_and_get(self.__attributes, "date_format", date_format),
            _add_to_dict_and_get(self.__attributes, "number_format", number_format),
        )
        if columns is not None:
            self.__update_col_desc_from_indexed(columns, "nan_value")
            self.__update_col_desc_from_indexed(columns, "width")
            group_by = self.__get_name_indexed_property("group_by")
            for k, v in group_by.items():
                if _is_boolean_true(v):
                    if col_desc := next((x for x in columns.values() if x["dfid"] == k), None):
                        col_desc["groupBy"] = True
                    else:
                        warnings.warn(f"{self.__element_name} group_by[{k}] is not in the list of displayed columns")
            apply = self.__get_name_indexed_property("apply")
            for k, v in apply.items():  # pragma: no cover
                if col_desc := next((x for x in columns.values() if x["dfid"] == k), None):
                    if callable(v):
                        value = self.__hashes.get(f"apply[{k}]")
                    elif isinstance(v, str):
                        value = v.strip()
                    else:
                        warnings.warn(f"{self.__element_name} apply[{k}] should be a user or predefined function")
                        value = None
                    if value:
                        col_desc["apply"] = value
                else:
                    warnings.warn(f"{self.__element_name} apply[{k}] is not in the list of displayed columns")
            if line_style := self.__attributes.get("style"):
                if callable(line_style):
                    value = self.__hashes.get("style")
                elif isinstance(line_style, str):
                    value = line_style.strip()
                else:
                    value = None
                if value in col_types.keys():
                    warnings.warn(f"{self.__element_name} style={value} cannot be a column's name")
                elif value:
                    self.set_attribute("lineStyle", value)
            styles = self.__get_name_indexed_property("style")
            for k, v in styles.items():  # pragma: no cover
                if col_desc := next((x for x in columns.values() if x["dfid"] == k), None):
                    if callable(v):
                        value = self.__hashes.get(f"style[{k}]")
                    elif isinstance(v, str):
                        value = v.strip()
                    else:
                        value = None
                    if value in col_types.keys():
                        warnings.warn(f"{self.__element_name} style[{k}]={value} cannot be a column's name")
                    elif value:
                        col_desc["style"] = value
                else:
                    warnings.warn(f"{self.__element_name} style[{k}] is not in the list of displayed columns")
            self.__attributes["columns"] = columns
            self.__set_json_attribute("columns", columns)
        return self

    def __check_dict(self, values: t.List[t.Any], indexes: t.Tuple[int], names: t.Tuple[str]) -> None:
        for index in indexes:
            if values[index] is not None and not isinstance(values[index], (dict, _MapDict)):
                warnings.warn(f"{self.__element_name} {names[index]} should be a dict")
                values[index] = None

    def get_chart_config(self, default_type="scatter", default_mode="lines+markers"):
        names = (
            "x",
            "y",
            "z",
            "label",
            "text",
            "mode",
            "type",
            "color",
            "xaxis",
            "yaxis",
            "selected_color",
            "marker",
            "selected_marker",
            "orientation",
            "name",
            "line",
            "text_anchor",
            "options",
        )
        trace = self.__get_multiple_indexed_attributes(names)
        if not trace[5]:
            # mode
            trace[5] = default_mode
        trace[6] = str(trace[6]).strip().lower() if trace[6] else default_type
        if not trace[8]:
            # xaxis
            trace[8] = "x"
        if not trace[9]:
            # yaxis
            trace[9] = "y"
        self.__check_dict(trace, (11, 12, 17), names)
        traces = []
        idx = 1
        indexed_trace = self.__get_multiple_indexed_attributes(names, idx)
        if len([x for x in indexed_trace if x]):
            while len([x for x in indexed_trace if x]):
                self.__check_dict(indexed_trace, (11, 12, 17), names)
                traces.append([x or trace[i] for i, x in enumerate(indexed_trace)])
                idx += 1
                indexed_trace = self.__get_multiple_indexed_attributes(names, idx)
        else:
            traces.append(trace)

        # read column definitions
        data = self.__attributes.get("data")
        data_hash = self.__hashes.get("data", "")
        col_types = self.__gui._accessors._get_col_types(data_hash, _TaipyData(data, data_hash))

        # add trace for non used indexed columns
        max_idx = max(_get_idx_from_col(c) for c in col_types.keys())
        traces.extend([x if i > 4 else None for i, x in enumerate(traces[0])] for _ in range(len(traces), max_idx + 1))

        # configure columns
        columns = set()
        for trace in traces:
            columns.update([t for t in trace[:5] if t])
        columns = _get_columns_dict(data, list(columns), col_types)
        # set default columns if not defined
        icols = [[c for c in [_get_col_from_indexed(c, i) for c in columns.keys()] if c] for i in range(len(traces))]

        for i, tr in enumerate(traces):
            if not tr[0] or tr[6] in _Builder.__ONE_COLUMN_CHART or not tr[1]:
                traces[i] = tuple(
                    v or (icols[i].pop(0) if j < 3 and j < len(icols[i]) else v) for j, v in enumerate(tr)
                )

        if columns is not None:
            self.__attributes["columns"] = columns
            reverse_cols = {cd["dfid"]: c for c, cd in columns.items()}

            ret_dict = {
                "columns": columns,
                "labels": [reverse_cols.get(t[3], (t[3] or "")) for t in traces],
                "texts": [reverse_cols.get(t[4], (t[4] or None)) for t in traces],
                "modes": [t[5] for t in traces],
                "types": [t[6] for t in traces],
                "xaxis": [t[8] for t in traces],
                "yaxis": [t[9] for t in traces],
                "markers": [t[11] or ({"color": t[7]} if t[7] else None) for t in traces],
                "selectedMarkers": [t[12] or ({"color": t[10]} if t[10] else None) for t in traces],
                "traces": [[reverse_cols.get(c, c) for c in [t[0], t[1], t[2]]] for t in traces],
                "orientations": [t[13] for t in traces],
                "names": [t[14] for t in traces],
                "lines": [t[15] if isinstance(t[15], dict) else {"dash": t[15]} for t in traces],
                "textAnchors": [t[16] for t in traces],
                "options": [t[17] for t in traces],
            }

            self.__set_json_attribute("config", ret_dict)
            self.set_chart_selected(max=len(traces))
        return self

    def set_chart_layout(self):
        if layout := self.__attributes.get("layout"):
            if isinstance(layout, (dict, _MapDict)):
                self.__set_json_attribute("layout", layout)
            else:
                warnings.warn(f"Chart control: layout attribute should be a dict\n'{str(layout)}'")

    def set_string_with_check(self, var_name: str, values: t.List[str], default_value: t.Optional[str] = None):
        value = self.__attributes.get(var_name, default_value)
        if value is not None:
            value = str(value).lower()
            self.__attributes[var_name] = value
            if value not in values:
                warnings.warn(f"{self.__element_name} {var_name}={value} should be in {values}")
            else:
                self.__set_string_attribute(var_name, default_value)
        return self

    def __set_list_attribute(self, name: str, hash_name: t.Optional[str], val: t.Any, elt_type: t.Type, dynamic=True) -> t.List[str]:
        if not hash_name and isinstance(val, str):
            val = [elt_type(t.strip()) for t in val.split(";")]
        if isinstance(val, list):
            if hash_name and dynamic:
                self.__set_react_attribute(name, hash_name)
                return [f"{name}={hash_name}"]
            else:
                self.__set_json_attribute(name, val)
        elif val is not None:
            warnings.warn(f"{self.__element_name} {name} should be a list of {elt_type}")
        return []

    def set_chart_selected(self, max=0):
        name = "selected"
        default_sel = self.__attributes.get(name)
        idx = 1
        name_idx = f"{name}[{idx}]"
        sel = self.__attributes.get(name_idx)
        while idx <= max:
            if sel is not None or default_sel is not None:
                self.__update_vars.extend(
                    self.__set_list_attribute(
                        f"{name}{idx - 1}",
                        self.__hashes.get(name_idx if sel is not None else name),
                        sel if sel is not None else default_sel,
                        int,
                    )
                )
            idx += 1
            name_idx = f"{name}[{idx}]"
            sel = self.__attributes.get(name_idx)

    def get_list_attribute(self, name: str, list_type: _AttributeType):
        varname = self.__hashes.get(name)
        if varname is None:
            list_val = self.__attributes.get(name)
            if isinstance(list_val, str):
                list_val = [s for s in list_val.split(";")]
            if isinstance(list_val, list):
                # TODO catch the cast exception
                if list_type.value == _AttributeType.number.value:
                    list_val = [int(v) for v in list_val]
                else:
                    list_val = [int(v) for v in list_val]
            else:
                if list_val is not None:
                    warnings.warn(f"{self.__element_name} {name} should be a list")
                list_val = []
            self.__set_react_attribute(_to_camel_case(name), list_val)
        else:
            self.__set_react_attribute(_to_camel_case(name), varname)
        return self

    def __set_class_names(self):
        classes = ["taipy-" + self.__control_type.replace("_", "-")]
        if cl := self.__attributes.get("class_name"):
            classes.append(str(cl))

        return self.set_attribute("className", " ".join(classes))

    def set_dataType(self):
        value = self.__attributes.get("value")
        return self.set_attribute("dataType", _get_data_type(value))

    def set_file_content(self, var_name: str = "content"):
        if hash_name := self.__hashes.get(var_name):
            self.__set_update_var_name(hash_name)
        else:
            warnings.warn("{self.element_name} {var_name} should be binded")
        return self

    def set_content(self, var_name: str = "content", image=True):
        content = self.__attributes.get(var_name)
        hash_name = self.__hashes.get(var_name)
        if content is None and hash_name is None:
            return self
        if hash_name:
            hash_name = self.__get_typed_hash_name(hash_name, _AttributeType.image if image else _AttributeType.content)
        value = self.__gui._get_content(hash_name or var_name, content, image)
        if hash_name:
            self.__set_react_attribute(
                var_name,
                _get_client_var_name(hash_name),
            )
        return self.set_attribute(_to_camel_case(f"default_{var_name}"), value)

    def set_lov(self, var_name="lov", property_name: t.Optional[str] = None):
        property_name = var_name if property_name is None else property_name
        self.__set_list_of_(f"default_{property_name}")
        if hash_name := self.__hashes.get(var_name):
            hash_name = self.__get_typed_hash_name(hash_name, _AttributeType.lov)
            self.__update_vars.append(f"{property_name}={hash_name}")
            self.__set_react_attribute(property_name, hash_name)
        return self

    def __set_dynamic_string_list(self, var_name: str, default_value: t.Any):
        hash_name = self.__hashes.get(var_name)
        loi = self.__attributes.get(var_name)
        if loi is None:
            loi = default_value
        if isinstance(loi, str):
            loi = [s.strip() for s in loi.split(";") if s.strip()]
        if isinstance(loi, list):
            self.__set_json_attribute(_to_camel_case(f"default_{var_name}"), loi)
        if hash_name:
            self.__update_vars.append(f"{var_name}={hash_name}")
            self.__set_react_attribute(var_name, hash_name)
        return self

    def __set_dynamic_number_attribute(self, var_name: str, default_value: t.Any):
        hash_name = self.__hashes.get(var_name)
        numVal = self.__attributes.get(var_name)
        if numVal is None:
            numVal = default_value
        if isinstance(numVal, str):
            try:
                numVal = float(numVal)
            except Exception as e:
                warnings.warn(f"{self.__element_name} {var_name} cannot be transformed into a number\n{e}")
                numVal = 0
        if isinstance(numVal, numbers.Number):
            self.__set_react_attribute(_to_camel_case(f"default_{var_name}"), numVal)
        elif numVal is not None:
            warnings.warn(f"{self.__element_name} {var_name} value is not not valid {numVal}")
        if hash_name:
            hash_name = self.__get_typed_hash_name(hash_name, _AttributeType.number)
            self.__update_vars.append(f"{var_name}={hash_name}")
            self.__set_react_attribute(var_name, hash_name)
        return self

    def __set_default_value(self, var_name: str, value: t.Optional[t.Any] = None, native_type: bool = False):
        if value is None:
            value = self.__attributes.get(var_name)
        default_var_name = _to_camel_case(f"default_{var_name}")
        if isinstance(value, (datetime, date, time)):
            self.set_attribute(default_var_name, _date_to_ISO(value))
        elif isinstance(value, str):
            self.set_attribute(default_var_name, value)
        elif native_type and isinstance(value, numbers.Number):
            self.__set_react_attribute(default_var_name, value)
        elif value is None:
            self.__set_react_attribute(default_var_name, "null")
        else:
            self.__set_json_attribute(default_var_name, value)
        return self

    def __set_update_var_name(self, hash_name: str):
        return self.set_attribute("updateVarName", hash_name)

    def set_value_and_default(
        self,
        var_name: t.Optional[str] = None,
        with_update=True,
        with_default=True,
        native_type=False,
        var_type: t.Optional[_AttributeType] = None,
        default_val: t.Any = None,
    ):
        var_name = self.__default_property_name if var_name is None else var_name
        if var_type == _AttributeType.number_or_lov_value:
            var_type = _AttributeType.lov_value if self.__attributes.get("lov") else _AttributeType.dynamic_number
            native_type = native_type if var_type == _AttributeType.dynamic_number else False
        if var_type == _AttributeType.dynamic_boolean:
            return self.set_attributes([(var_name, var_type, bool(default_val), with_update)])
        if hash_name := self.__hashes.get(var_name):
            hash_name = self.__get_typed_hash_name(hash_name, var_type)
            self.__set_react_attribute(
                var_name,
                _get_client_var_name(hash_name),
            )
            if with_update:
                self.__set_update_var_name(hash_name)
            if with_default:
                if native_type:
                    val = self.__attributes.get(var_name)
                    if native_type and isinstance(val, str):
                        with contextlib.suppress(Exception):
                            val = float(val)
                    self.__set_default_value(var_name, val, native_type=native_type)
                else:
                    self.__set_default_value(var_name)
        else:
            value = self.__attributes.get(var_name)
            if value is not None:
                if native_type:
                    if isinstance(value, str):
                        with contextlib.suppress(Exception):
                            value = float(value)
                    if isinstance(value, (int, float)):
                        return self.__set_react_attribute(var_name, value)
                self.set_attribute(var_name, value)
        return self

    def set_labels(self, var_name: str = "labels"):
        if value := self.__attributes.get(var_name):
            if _is_boolean_true(value):
                return self.__set_react_attribute(_to_camel_case(var_name), True)
            return self.__set_dict_attribute(var_name)
        return self

    def set_partial(self):
        if self.__control_type not in _Builder.__BLOCK_CONTROLS:
            return self
        if partial := self.__attributes.get("partial"):
            page = self.__attributes.get("page")
            if page:
                warnings.warn(
                    f"{self.__element_name} control: page and partial should not be defined at the same time."
                )
            if isinstance(partial, Partial):
                self.__attributes["page"] = partial._route
                self.__set_react_attribute("partial", partial._route)
        return self

    def set_propagate(self):
        val = self.__get_boolean_attribute("propagate", self.__gui._config.config.get("propagate"))
        if not val:
            return self.__set_boolean_attribute("propagate", False)
        return self

    def set_refresh_on_update(self):
        if self.__update_vars:
            self.set_attribute("updateVars", ";".join(self.__update_vars))
        return self

    def set_table_pagesize_options(self, default_size=[50, 100, 500]):
        page_size_options = self.__attributes.get("page_size_options", default_size)
        if isinstance(page_size_options, str):
            try:
                page_size_options = [int(s.strip()) for s in page_size_options.split(";")]
            except Exception as e:
                warnings.warn(f"{self.__element_name} page_size_options: invalid value {page_size_options}\n{e}")
        if isinstance(page_size_options, list):
            self.__set_json_attribute("pageSizeOptions", page_size_options)
        else:
            warnings.warn(f"{self.__element_name} page_size_options should be a list")
        return self

    def set_type(self, type_name: str):
        self.type_name = type_name
        self.set_attribute("type", type_name)
        return self

    def set_kind(self):
        theme = self.__attributes.get("theme", False)
        if theme:
            self.set_attribute("kind", "theme")
        return self

    def __get_typed_hash_name(self, hash_name: str, var_type: t.Optional[_AttributeType]) -> str:
        if taipy_type := _get_taipy_type(var_type):
            expr = self.__gui._get_expr_from_hash(hash_name)
            hash_name = self.__gui._evaluate_bind_holder(taipy_type, expr)
        return hash_name

    def __set_dynamic_bool_attribute(self, name: str, def_val: t.Any, with_update: bool, update_main=True):
        hash_name = self.__hashes.get(name)
        val = self.__get_boolean_attribute(name, def_val)
        default_name = f"default_{name}" if hash_name is not None else name
        if val != def_val:
            self.__set_boolean_attribute(default_name, val)
        if hash_name is not None:
            hash_name = self.__get_typed_hash_name(hash_name, _AttributeType.dynamic_boolean)
            self.__set_react_attribute(_to_camel_case(name), _get_client_var_name(hash_name))
            if with_update:
                if update_main:
                    self.__set_update_var_name(hash_name)
                else:
                    self.__update_vars.append(f"{name}={hash_name}")

    def set_attributes(self, attributes: t.List[tuple]):
        for attr in attributes:
            if not isinstance(attr, tuple):
                attr = (attr,)
            var_type = _get_tuple_val(attr, 1, _AttributeType.string)
            if var_type == _AttributeType.boolean:
                def_val = _get_tuple_val(attr, 2, False)
                val = self.__get_boolean_attribute(attr[0], def_val)
                if val != def_val:
                    self.__set_boolean_attribute(attr[0], val)
            elif var_type == _AttributeType.dynamic_boolean:
                self.__set_dynamic_bool_attribute(
                    attr[0], _get_tuple_val(attr, 2, False), _get_tuple_val(attr, 3, False)
                )
            elif var_type == _AttributeType.number:
                self.__set_number_attribute(attr[0], _get_tuple_val(attr, 2, None))
            elif var_type == _AttributeType.dynamic_number:
                self.__set_dynamic_number_attribute(attr[0], _get_tuple_val(attr, 2, None))
            elif var_type == _AttributeType.string:
                self.__set_string_attribute(attr[0], _get_tuple_val(attr, 2, None), _get_tuple_val(attr, 3, True))
            elif var_type == _AttributeType.dynamic_string:
                self.__set_dynamic_string_attribute(
                    attr[0], _get_tuple_val(attr, 2, None), _get_tuple_val(attr, 3, False)
                )
            elif var_type == _AttributeType.string_list:
                self.__set_list_attribute(attr[0], self.__hashes.get(
                    attr[0]), self.__attributes.get(attr[0]), str, False)
            elif var_type == _AttributeType.function:
                self.__set_function_attribute(attr[0], _get_tuple_val(attr, 2, None), _get_tuple_val(attr, 3, True))
            elif var_type == _AttributeType.react:
                self.__set_react_attribute(_to_camel_case(attr[0]), _get_tuple_val(attr, 2, None))
            elif var_type == _AttributeType.string_or_number:
                self.__set_string_or_number_attribute(attr[0], _get_tuple_val(attr, 2, None))
            elif var_type == _AttributeType.dict:
                self.__set_dict_attribute(attr[0])
            elif var_type == _AttributeType.dynamic_list:
                self.__set_dynamic_string_list(attr[0], _get_tuple_val(attr, 2, None))
            elif var_type == _AttributeType.boolean_or_list:
                if _is_boolean(self.__attributes.get(attr[0])):
                    self.__set_dynamic_bool_attribute(attr[0], _get_tuple_val(attr, 2, False), True, update_main=False)
                else:
                    self.__set_dynamic_string_list(attr[0], _get_tuple_val(attr, 2, None))
        return self

    def set_attribute(self, name, value):
        if name.startswith("on"):
            name = f"tp_{name}"
        self.el.set(name, value)
        return self

    def build_to_string(self):
        el_str = str(etree.tostring(self.el, encoding="utf8").decode("utf8"))
        el_str = el_str.replace("<?xml version='1.0' encoding='utf8'?>\n", "")
        el_str = el_str.replace("/>", ">")
        return el_str, self.__element_name



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
import warnings
from datetime import date, datetime, time

from dateutil import parser
from pytz import utc


def _date_to_ISO(date_val: t.Union[datetime, date, time]) -> str:
    if isinstance(date_val, datetime):
        # return date.isoformat() + 'Z', if possible
        try:
            return date_val.astimezone(utc).isoformat()
        except Exception as e:
            # astimezone() fails on Windows for pre-epoch times
            # See https://bugs.python.org/issue36759
            warnings.warn(f"There is some problems with date parsing to ISO!\n{e}")
    return date_val.isoformat()


def _ISO_to_date(date_str: str) -> datetime:
    # return datetime.strptime(date_str, '%Y-%m-%dT%H:%M:%S.%fZ')
    # return datetime.fromisoformat(date_str)
    return parser.parse(date_str)



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re
import typing as t
from datetime import datetime

from ..types import _AttributeType
from .builder import _Builder


class _Factory:

    DEFAULT_CONTROL = "text"

    __CONTROL_DEFAULT_PROP_NAME = {
        "button": "label",
        "chart": "data",
        "content": "value",
        "date": "date",
        "dialog": "open",
        "expandable": "title",
        "file_download": "content",
        "file_selector": "content",
        "image": "content",
        "indicator": "display",
        "input": "value",
        "layout": "columns",
        "menu": "lov",
        "navbar": "value",
        "number": "value",
        "pane": "open",
        "part": "render",
        "selector": "value",
        "slider": "value",
        "status": "value",
        "table": "data",
        "text": "value",
        "toggle": "value",
        "tree": "value",
    }

    __TEXT_ANCHORS = ["bottom", "top", "left", "right"]
    __TEXT_ANCHOR_NONE = "none"

    CONTROL_BUILDERS = {
        "button": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="Button",
            attributes=attrs,
        )
        .set_value_and_default(with_update=False)
        .set_attributes(
            [
                ("id",),
                ("on_action", _AttributeType.function, ""),
                ("active", _AttributeType.dynamic_boolean, True),
                ("hover_text", _AttributeType.dynamic_string),
            ]
        ),
        "chart": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="Chart",
            attributes=attrs,
        )
        .set_value_and_default(with_default=False, var_type=_AttributeType.data)
        .set_attributes(
            [
                ("id",),
                ("title",),
                ("width", _AttributeType.string_or_number),
                ("height", _AttributeType.string_or_number),
                ("layout", _AttributeType.dict),
                ("plot_config", _AttributeType.dict),
                ("on_range_change", _AttributeType.function),
                ("active", _AttributeType.dynamic_boolean, True),
                ("limit_rows", _AttributeType.boolean),
                ("render", _AttributeType.dynamic_boolean, True),
                ("hover_text", _AttributeType.dynamic_string),
                ("on_change", ),
            ]
        )
        .get_chart_config("scatter", "lines+markers")
        .set_propagate()
        .set_refresh_on_update(),
        "content": lambda gui, control_type, attrs: _Builder(
            gui=gui, control_type=control_type, element_name="PageContent", attributes=attrs
        ),
        "date": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="DateSelector",
            attributes=attrs,
            default_value=datetime.fromtimestamp(0),
        )
        .set_value_and_default(var_type=_AttributeType.date)
        .set_attributes(
            [
                ("with_time", _AttributeType.boolean),
                ("id",),
                ("active", _AttributeType.dynamic_boolean, True),
                ("editable", _AttributeType.dynamic_boolean, True),
                ("hover_text", _AttributeType.dynamic_string),
                ("on_change", ),
            ]
        )
        .set_propagate(),
        "dialog": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="Dialog",
            attributes=attrs,
        )
        .set_value_and_default(var_type=_AttributeType.dynamic_boolean)
        .set_partial()  # partial should be set before page
        .set_attributes(
            [
                ("id",),
                ("page",),
                ("title",),
                ("on_action", _AttributeType.function),
                ("close_label", _AttributeType.string),
                ("labels", _AttributeType.string_list),
                ("active", _AttributeType.dynamic_boolean, True),
                ("width", _AttributeType.string_or_number),
                ("height", _AttributeType.string_or_number),
                ("hover_text", _AttributeType.dynamic_string),
            ]
        )
        .set_propagate(),
        "expandable": lambda gui, control_type, attrs: _Builder(
            gui=gui, control_type=control_type, element_name="Expandable", attributes=attrs, default_value=None
        )
        .set_value_and_default()
        .set_partial()  # partial should be set before page
        .set_attributes(
            [
                ("id",),
                ("page",),
                ("expanded", _AttributeType.dynamic_boolean, True),
                ("hover_text", _AttributeType.dynamic_string),
            ]
        ),
        "file_download": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="FileDownload",
            attributes=attrs,
        )
        .set_value_and_default(var_name="label", with_update=False)
        .set_content("content", image=False)
        .set_attributes(
            [
                ("id",),
                ("on_action", _AttributeType.function),
                ("active", _AttributeType.dynamic_boolean, True),
                ("render", _AttributeType.dynamic_boolean, True),
                ("auto", _AttributeType.boolean, False),
                ("bypass_preview", _AttributeType.boolean, True),
                ("name",),
                ("hover_text", _AttributeType.dynamic_string),
            ]
        ),
        "file_selector": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="FileSelector",
            attributes=attrs,
        )
        .set_value_and_default(var_name="label", with_update=False)
        .set_file_content()
        .set_attributes(
            [
                ("id",),
                ("on_action", _AttributeType.function),
                ("active", _AttributeType.dynamic_boolean, True),
                ("multiple", _AttributeType.boolean, False),
                ("extensions",),
                ("drop_message",),
                ("hover_text", _AttributeType.dynamic_string),
            ]
        ),
        "image": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="Image",
            attributes=attrs,
        )
        .set_value_and_default(var_name="label", with_update=False)
        .set_content("content")
        .set_attributes(
            [
                ("id",),
                ("on_action", _AttributeType.function, ""),
                ("active", _AttributeType.dynamic_boolean, True),
                ("width",),
                ("height",),
                ("hover_text", _AttributeType.dynamic_string),
            ]
        ),
        "indicator": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="Indicator",
            attributes=attrs,
        )
        .set_value_and_default(with_update=False, native_type=True)
        .set_attributes(
            [
                ("id",),
                ("min", _AttributeType.number),
                ("max", _AttributeType.number),
                ("value", _AttributeType.dynamic_number),
                ("format",),
                ("orientation"),
                ("hover_text", _AttributeType.dynamic_string),
            ]
        ),
        "input": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="Input",
            attributes=attrs,
        )
        .set_type("text")
        .set_value_and_default()
        .set_propagate()
        .set_attributes(
            [
                ("id",),
                ("active", _AttributeType.dynamic_boolean, True),
                ("hover_text", _AttributeType.dynamic_string),
                ("on_change", ),
            ]
        ),
        "layout": lambda gui, control_type, attrs: _Builder(
            gui=gui, control_type=control_type, element_name="Layout", attributes=attrs, default_value=None
        )
        .set_value_and_default(with_default=False)
        .set_attributes(
            [
                ("id",),
                ("columns[mobile]",),
                ("gap",),
            ]
        ),
        "menu": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="MenuCtl",
            attributes=attrs,
        )
        .get_adapter("lov")  # need to be called before set_lov
        .set_lov()
        .set_attributes(
            [
                ("id",),
                ("active", _AttributeType.dynamic_boolean, True),
                ("label"),
                ("width"),
                ("width[mobile]",),
                ("on_action", _AttributeType.function, "on_menu_action"),
                ("inactive_ids", _AttributeType.dynamic_list),
                ("hover_text", _AttributeType.dynamic_string),
            ]
        )
        .set_refresh_on_update()
        .set_propagate(),
        "navbar": lambda gui, control_type, attrs: _Builder(
            gui=gui, control_type=control_type, element_name="NavBar", attributes=attrs, default_value=None
        )
        .get_adapter("lov", multi_selection=False)  # need to be called before set_lov
        .set_lov()
        .set_attributes(
            [
                ("id",),
                ("active", _AttributeType.dynamic_boolean, True),
                ("hover_text", _AttributeType.dynamic_string),
            ]
        ),
        "number": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="Input",
            attributes=attrs,
            default_value=0,
        )
        .set_type("number")
        .set_value_and_default()
        .set_propagate()
        .set_attributes(
            [
                ("id",),
                ("active", _AttributeType.dynamic_boolean, True),
                ("hover_text", _AttributeType.dynamic_string),
                ("on_change", ),
            ]
        ),
        "pane": lambda gui, control_type, attrs: _Builder(
            gui=gui, control_type=control_type, element_name="Pane", attributes=attrs, default_value=None
        )
        .set_value_and_default(var_type=_AttributeType.dynamic_boolean)
        .set_partial()  # partial should be set before page
        .set_attributes(
            [
                ("id",),
                ("page",),
                ("anchor", _AttributeType.string, "left"),
                ("on_close", _AttributeType.function),
                ("persistent", _AttributeType.boolean, False),
                ("active", _AttributeType.dynamic_boolean, True),
                ("width", _AttributeType.string_or_number, "30vw"),
                ("height", _AttributeType.string_or_number, "30vh"),
                ("hover_text", _AttributeType.dynamic_string),
                ("on_change", ),
            ]
        )
        .set_propagate(),
        "part": lambda gui, control_type, attrs: _Builder(
            gui=gui, control_type=control_type, element_name="Part", attributes=attrs, default_value=None
        )
        .set_value_and_default(with_update=False, var_type=_AttributeType.dynamic_boolean, default_val=True)
        .set_partial()  # partial should be set before page
        .set_attributes(
            [
                ("id",),
                ("page",),
            ]
        ),
        "selector": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="Selector",
            attributes=attrs,
        )
        .set_value_and_default(with_default=False, var_type=_AttributeType.lov_value)
        .get_adapter("lov")  # need to be called before set_lov
        .set_lov()
        .set_attributes(
            [
                ("active", _AttributeType.dynamic_boolean, True),
                ("dropdown", _AttributeType.boolean, False),
                ("filter", _AttributeType.boolean),
                ("height", _AttributeType.string_or_number),
                ("hover_text", _AttributeType.dynamic_string),
                ("id",),
                ("value_by_id", _AttributeType.boolean),
                ("multiple", _AttributeType.boolean),
                ("width", _AttributeType.string_or_number),
                ("on_change", ),
            ]
        )
        .set_refresh_on_update()
        .set_propagate(),
        "slider": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="Slider",
            attributes=attrs,
            default_value=0,
        )
        .set_value_and_default(native_type=True, var_type=_AttributeType.number_or_lov_value)
        .set_attributes(
            [
                ("active", _AttributeType.dynamic_boolean, True),
                ("height"),
                ("hover_text", _AttributeType.dynamic_string),
                ("id",),
                ("value_by_id", _AttributeType.boolean),
                ("max", _AttributeType.number, 100),
                ("min", _AttributeType.number, 0),
                ("orientation"),
                ("width", _AttributeType.string, "300px"),
                ("on_change", ),
            ]
        )
        .get_adapter("lov")  # need to be called before set_lov
        .set_lov()
        .set_labels()
        .set_string_with_check("text_anchor", _Factory.__TEXT_ANCHORS + [_Factory.__TEXT_ANCHOR_NONE], "bottom")
        .set_refresh_on_update()
        .set_propagate(),
        "status": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="Status",
            attributes=attrs,
        )
        .set_value_and_default(with_update=False)
        .set_attributes(
            [
                ("id",),
                ("without_close", _AttributeType.boolean, False),
                ("hover_text", _AttributeType.dynamic_string),
            ]
        )
        .set_refresh_on_update(),
        "table": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="Table",
            attributes=attrs,
        )
        .set_value_and_default(with_default=False, var_type=_AttributeType.data)
        .get_dataframe_attributes()
        .set_attributes(
            [
                ("page_size", _AttributeType.react, 100),
                ("allow_all_rows", _AttributeType.boolean),
                ("show_all", _AttributeType.boolean),
                ("auto_loading", _AttributeType.boolean),
                ("width", _AttributeType.string_or_number, "100vw"),
                ("height", _AttributeType.string_or_number, "80vh"),
                ("id",),
                ("active", _AttributeType.dynamic_boolean, True),
                ("editable", _AttributeType.dynamic_boolean, True),
                ("on_edit", _AttributeType.function),
                ("on_delete", _AttributeType.function),
                ("on_add", _AttributeType.function),
                ("nan_value",),
                ("hover_text", _AttributeType.dynamic_string),
            ]
        )
        .set_propagate()
        .get_list_attribute("selected", _AttributeType.number)
        .set_refresh_on_update()
        .set_table_pagesize_options(),
        "text": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="Field",
            attributes=attrs,
        )
        .set_value_and_default(with_update=False)
        .set_dataType()
        .set_attributes(
            [
                ("format",),
                ("id",),
                ("hover_text", _AttributeType.dynamic_string),
            ]
        ),
        "toggle": lambda gui, control_type, attrs: _Builder(
            gui=gui, control_type=control_type, element_name="Toggle", attributes=attrs, default_value=None
        )
        .set_value_and_default(with_default=False, var_type=_AttributeType.lov_value)
        .get_adapter("lov", multi_selection=False)  # need to be called before set_lov
        .set_lov()
        .set_attributes(
            [
                ("active", _AttributeType.dynamic_boolean, True),
                ("hover_text", _AttributeType.dynamic_string),
                ("id",),
                ("label",),
                ("value_by_id", _AttributeType.boolean),
                ("unselected_value", _AttributeType.string, ""),
                ("on_change", ),
            ]
        )
        .set_kind()
        .set_refresh_on_update()
        .set_propagate(),
        "tree": lambda gui, control_type, attrs: _Builder(
            gui=gui,
            control_type=control_type,
            element_name="TreeView",
            attributes=attrs,
        )
        .set_value_and_default(with_default=False, var_type=_AttributeType.lov_value)
        .get_adapter("lov")  # need to be called before set_lov
        .set_lov()
        .set_attributes(
            [
                ("active", _AttributeType.dynamic_boolean, True),
                ("expanded", _AttributeType.boolean_or_list, True),
                ("filter", _AttributeType.boolean),
                ("hover_text", _AttributeType.dynamic_string),
                ("height", _AttributeType.string_or_number),
                ("id",),
                ("value_by_id", _AttributeType.boolean),
                ("multiple", _AttributeType.boolean),
                ("width", _AttributeType.string_or_number),
                ("on_change", ),
            ]
        )
        .set_refresh_on_update()
        .set_propagate(),
    }

    # TODO: process \" in property value
    _PROPERTY_RE = re.compile(r"\s+([a-zA-Z][\.a-zA-Z_$0-9]*(?:\[(?:.*?)\])?)=\"((?:(?:(?<=\\)\")|[^\"])*)\"")

    @staticmethod
    def get_default_property_name(control_name: str) -> t.Optional[str]:
        return _Factory.__CONTROL_DEFAULT_PROP_NAME.get(control_name.split(".", 1)[0])



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re
import typing as t
import warnings
from typing import Any, List, Tuple

from markdown.preprocessors import Preprocessor as MdPreprocessor

from ..builder import _Builder
from .factory import _MarkdownFactory


class _Preprocessor(MdPreprocessor):
    # ----------------------------------------------------------------------
    # Finds, in the Markdown text, control declaration constructs:
    #     <|<some value>|>
    # or
    #     <|<some value>|<control_type>|>
    # or
    #     <|<some value>|<control_type>|<prop_name[=propvalue]>>
    # or
    #     <|<control_type>|<prop_name[=propvalue]>>
    #
    # These constructs are converted a fragment that the ControlPattern
    # processes to create the components that get generated.
    #     <control_type> prop_name="prop_value" ...
    # Note that if a value is provided before the control_type, it is set
    # as the default property value for that control type.
    # The default control type is 'text'.
    # ----------------------------------------------------------------------
    # Control in Markdown
    __CONTROL_RE = re.compile(r"<\|(.*?)\|>")
    # Opening tag
    __OPENING_TAG_RE = re.compile(r"<([0-9a-zA-Z\_\.]*)\|((?:(?!\|>).)*)\s*$")
    # Closing tag
    __CLOSING_TAG_RE = re.compile(r"^\s*\|([0-9a-zA-Z\_\.]*)>")
    # Link in Markdown
    __LINK_RE = re.compile(r"(\[[^\]]*?\]\([^\)]*?\))")
    # Split properties and control type
    __SPLIT_RE = re.compile(r"(?<!\\\\)\|")
    # Property syntax: '<prop_name>[=<prop_value>]'
    #   If <prop_value> is omitted:
    #     '<prop_name>' is equivalent to '<prop_name>=true'
    #     'not <prop_name>' is equivalent to '<prop_name>=false'
    #       'not', 'dont', 'don't' are equivalent in this context
    #  Note 1: 'not <prop_name>=<prop_value>' is an invalid syntax
    #  Note 2: Space characters after the equal sign are significative
    __PROPERTY_RE = re.compile(r"((?:don'?t|not)\s+)?([a-zA-Z][\.a-zA-Z_$0-9]*(?:\[(?:.*?)\])?)\s*(?:=(.*))?$")

    @staticmethod
    def extend(md, gui, priority):
        instance = _Preprocessor(md)
        md.preprocessors.register(instance, "taipy", priority)
        instance._gui = gui

    def _make_prop_pair(self, prop_name: str, prop_value: str) -> Tuple[str, str]:
        # Un-escape pipe character in property value
        return (prop_name, prop_value.replace("\\|", "|"))

    def run(self, lines: List[str]) -> List[str]:
        new_lines = []
        tag_queue = []
        for line_count, line in enumerate(lines, start=1):
            new_line = ""
            last_index = 0
            # Opening tags
            m = _Preprocessor.__OPENING_TAG_RE.search(line)
            if m is not None:
                tag = "part"
                properties: List[Tuple[str, str]] = []
                if m.group(2):
                    tag, properties = self._process_control(m.group(2), line_count)
                if tag in _MarkdownFactory._TAIPY_BLOCK_TAGS:
                    tag_queue.append((tag, line_count, m.group(1) or None))
                    new_line_delimeter = "\n" if line.startswith("<|") else "\n\n"
                    line = line[: m.start()] + new_line_delimeter + _MarkdownFactory._TAIPY_START + tag + ".start"
                    for property in properties:
                        prop_value = property[1].replace('"', '\\"')
                        line += f' {property[0]}="{prop_value}"'
                    line += _MarkdownFactory._TAIPY_END + new_line_delimeter
                else:
                    warnings.warn(f"Invalid tag name '{tag}' in line {line_count}")
            # Other controls
            for m in _Preprocessor.__CONTROL_RE.finditer(line):
                control_name, properties = self._process_control(m.group(1), line_count)
                new_line += line[last_index : m.start()]
                control_text = _MarkdownFactory._TAIPY_START + control_name
                for property in properties:
                    prop_value = property[1].replace('"', '\\"')
                    control_text += f' {property[0]}="{prop_value}"'
                control_text += _MarkdownFactory._TAIPY_END
                new_line += control_text
                last_index = m.end()
            new_line = line if last_index == 0 else new_line + line[last_index:]
            # Add key attribute to links
            line = new_line
            new_line = ""
            last_index = 0
            for m in _Preprocessor.__LINK_RE.finditer(line):
                new_line += line[last_index : m.end()]
                new_line += "{: key=" + _Builder._get_key("link") + "}"
                last_index = m.end()
            new_line = line if last_index == 0 else new_line + line[last_index:]
            # Look for a closing tag
            m = _Preprocessor.__CLOSING_TAG_RE.search(new_line)
            if m is not None:
                if len(tag_queue):
                    open_tag, open_tag_line_count, open_tag_identifier = tag_queue.pop()
                    close_tag_identifier = m.group(1)
                    if close_tag_identifier and not open_tag_identifier:
                        warnings.warn(
                            f"Missing opening '{open_tag}' tag identifier '{close_tag_identifier}' in line {open_tag_line_count}"
                        )
                    if open_tag_identifier and not close_tag_identifier:
                        warnings.warn(
                            f"Missing closing '{open_tag}' tag identifier '{open_tag_identifier}' in line {line_count}"
                        )
                    if close_tag_identifier and open_tag_identifier and close_tag_identifier != open_tag_identifier:
                        warnings.warn(
                            f"Unmatched '{open_tag}' tag identifier in line {open_tag_line_count} and line {line_count}"
                        )
                    new_line = (
                        new_line[: m.start()]
                        + _MarkdownFactory._TAIPY_START
                        + open_tag
                        + ".end"
                        + _MarkdownFactory._TAIPY_END
                        + "\n"
                        + new_line[m.end() :]
                    )
                else:
                    new_line = (
                        new_line[: m.start()]
                        + f"<div>No matching opened tag on line {line_count}</div>"
                        + new_line[m.end() :]
                    )
                    warnings.warn(f"Line {line_count} has an unmatched closing tag")
            # append the new line
            new_lines.append(new_line)
        # Issue #337: add an empty string at the beginning of new_lines list if there is not one
        # so that markdown extension would be able to render properly
        if new_lines and new_lines[0] != "":
            new_lines.insert(0, "")
        # Check for tags left unclosed (but close them anyway)
        for tag, line_no, _ in tag_queue:
            new_lines.append(_MarkdownFactory._TAIPY_START + tag + ".end" + _MarkdownFactory._TAIPY_END)
            warnings.warn(f"Opened tag {tag} in line {line_no} is not closed")
        return new_lines

    def _process_control(self, prop_string: str, line_count: int) -> Tuple[str, List[Tuple[str, str]]]:
        fragments = [f for f in _Preprocessor.__SPLIT_RE.split(prop_string) if f]
        control_name = None
        default_prop_name = None
        default_prop_value = None
        properties: List[Tuple[str, Any]] = []
        for fragment in fragments:
            if control_name is None and _MarkdownFactory.get_default_property_name(fragment):
                control_name = fragment
            elif control_name is None and default_prop_value is None:
                default_prop_value = fragment
            elif prop_match := _Preprocessor.__PROPERTY_RE.match(fragment):
                not_prefix = prop_match.group(1)
                prop_name = prop_match.group(2)
                val = prop_match.group(3)
                if not_prefix and val:
                    warnings.warn(f"Negated property {prop_name} value ignored at {line_count}")
                prop_value = "True"
                if not_prefix:
                    prop_value = "False"
                elif val:
                    prop_value = val
                properties.append(self._make_prop_pair(prop_name, prop_value))
            else:
                warnings.warn(f"Bad Taipy property format at line {line_count}: '{fragment}'")

        if control_name is None:
            control_name = _MarkdownFactory.DEFAULT_CONTROL
        if default_prop_value is not None:
            default_prop_name = _MarkdownFactory.get_default_property_name(control_name)
            # Set property only if it is not already defined
            if default_prop_name and default_prop_name not in [x[0] for x in properties]:
                properties.insert(0, self._make_prop_pair(default_prop_name, default_prop_value))
        return control_name, properties



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from ._attributes import (
    _delscopeattr,
    _getscopeattr,
    _getscopeattr_drill,
    _hasscopeattr,
    _setscopeattr,
    _setscopeattr_drill,
)
from ._map_dict import _MapDict
from .boolean import _is_boolean, _is_boolean_true
from .clientvarname import _get_client_var_name
from .datatype import _get_data_type
from .date import _date_to_ISO, _ISO_to_date
from .expr_var_name import _get_expr_var_name
from .filename import _get_non_existent_file_path
from .getdatecolstrname import _get_date_col_str_name, _RE_PD_TYPE
from .isnotebook import _is_in_notebook
from .killable_thread import _KillableThread
from .types import (
    _TaipyBase,
    _TaipyBool,
    _TaipyContent,
    _TaipyContentImage,
    _TaipyData,
    _TaipyDate,
    _TaipyLov,
    _TaipyLovValue,
    _TaipyNumber,
)
from .varnamefromcontent import _varname_from_content



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import base64
import pathlib
import tempfile
import typing as t
import warnings
from importlib import util
from pathlib import Path
from sys import getsizeof

from ..utils import _get_non_existent_file_path

_has_magic_module = False

if util.find_spec("magic"):
    import magic

    _has_magic_module = True


class _ContentAccessor:
    def __init__(self, data_url_max_size: int) -> None:
        self.__content_paths: t.Dict[str, pathlib.Path] = {}
        self.__url_is_image: t.Dict[str, bool] = {}
        self.__paths: t.Dict[pathlib.Path, str] = {}
        self.__data_url_max_size = data_url_max_size
        self.__temp_dir_path = Path(tempfile.gettempdir())

    def get_path(self, path: pathlib.Path) -> str:
        url_path = self.__paths.get(path)
        if url_path is None:
            self.__paths[path] = url_path = "taipyStatic" + str(len(self.__paths))
        return url_path

    def get_content_path(
        self, url_path: str, file_name: str, bypass: t.Union[str, None]
    ) -> t.Tuple[t.Union[pathlib.Path, None], bool]:
        content_path = self.__content_paths.get(url_path)
        if not content_path:
            return (None, True)
        return (content_path, bypass is not None or self.__url_is_image.get(f"{url_path}/{file_name}", False))

    def __get_mime_from_file(self, path: pathlib.Path):
        if _has_magic_module:
            try:
                return magic.from_file(str(path), mime=True)
            except Exception as e:
                warnings.warn(f"({path}) cannot read mime type.\n{e}")
        return None

    def get_info(self, var_name: str, value: t.Any, image: bool) -> t.Union[str, t.Tuple[str], t.Any]:  # noqa: C901
        if value is None:
            return ""
        newvalue = value
        mime = None
        if not isinstance(newvalue, (str, pathlib.Path)) and (
            getsizeof(newvalue) > self.__data_url_max_size or not _has_magic_module
        ):
            # write data to file
            file_name = "TaiPyContent.bin"
            if _has_magic_module:
                try:
                    mime = magic.from_buffer(value, mime=True)
                    file_name = "TaiPyContent." + mime.split("/")[-1]
                except Exception as e:
                    warnings.warn(f"{var_name} ({type(value)}) cannot be typed.\n{e}")
            file_path = _get_non_existent_file_path(self.__temp_dir_path, file_name)
            try:
                with open(file_path, "wb") as temp_file:
                    temp_file.write(value)
            except Exception as e:
                warnings.warn(f"{var_name} ({type(value)}) cannot be written to file {file_path}.\n{e}")
            newvalue = file_path
        if isinstance(newvalue, (str, pathlib.Path)):
            path = pathlib.Path(newvalue) if isinstance(newvalue, str) else newvalue
            if not path.is_file():
                return str(value)
            if image:
                if not mime:
                    mime = self.__get_mime_from_file(path)
                if mime and not mime.startswith("image"):
                    warnings.warn(f"{var_name} ({path}) is not an image: {mime}")
                    return f"Invalid content: {mime}"
            dir_path = path.resolve().parent
            url_path = self.get_path(dir_path)
            self.__content_paths[url_path] = dir_path
            file_url = f"{url_path}/{path.name}"
            self.__url_is_image[file_url] = image
            return (file_url,)
        elif _has_magic_module:
            try:
                mime = magic.from_buffer(value, mime=True)
                if not image or mime.startswith("image"):
                    return f"data:{mime};base64," + str(base64.b64encode(value), "utf-8")
                warnings.warn(f"{var_name} ({type(value)}) is not an image: {mime}")
                return f"Invalid content: {mime}"
            except Exception as e:
                warnings.warn(f"{var_name} ({type(value)}) cannot be base64 encoded.\n{e}")
                return "Cannot be base64 encoded"
        else:
            warnings.warn(
                "python-magic (and python-magic-bin for win) packages need to be installed if you want to process content as array of bytes."
            )
            return "Cannot guess content type"



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
import warnings

from .gui import Gui
from .state import State


def download(state: State, content: t.Any, name: t.Optional[str] = "", on_action: t.Optional[str] = ""):
    """Download content to the client.

    Arguments:
        state (State^): The current user state as received in any callback.
        content: File path or file content.
        name: File name for the content on the client browser (default to content name).
        on_action: Function called when the download starts.
    """
    if state and isinstance(state._gui, Gui):
        state._gui._download(content, name, on_action)
    else:
        warnings.warn("'download()' must be called in the context of a callback")


def notify(
    state: State,
    notification_type: str = "I",
    message: str = "",
    system_notification: t.Optional[bool] = None,
    duration: t.Optional[int] = None,
):
    """Send a notification to the user interface.

    Arguments:
        state (State^): The current user state as received in any callback.
        notification_type: The notification type. This can be one of _"success"_, _"info"_,
            _"warning"_, or _"error"_.<br/>
            To remove the last notification, set this parameter to the empty string.
        message: The text message to display.
        system_notification: If True, the system will also show the notification.
            If not specified or set to None, this parameter will use the value of
            _configuration[system_notification]_.
        duration: The time, in milliseconds, during which the notification is shown.
            If not specified or set to None, this parameter will use the value of
            _configuration[notification_duration]_.

    Note that you can also call this function with _notification_type_ set to the first letter
    or the alert type (ie setting _notification_type_ to "i" is equivalent to setting it to
    "info").

    If _system_notification_ is set to True, then the browser requests the system
    to display a notification as well. They usually appear in small windows that
    fly out of the system tray.<br/>
    The first time your browser is requested to show such a system notification for
    Taipy applications, you may be prompted to authorize the browser to do so. Please
    refer to your browser documentation for details on how to allow or prevent this
    feature.
    """
    if state and isinstance(state._gui, Gui):
        state._gui._notify(notification_type, message, system_notification, duration)
    else:
        warnings.warn("'notify()' must be called in the context of a callback")


def hold_control(
    state: State,
    callback: t.Optional[t.Union[str, t.Callable]] = None,
    message: t.Optional[str] = "Work in Progress...",
):
    """Hold the User Interface actions.

    When the User Interface is held, users cannot interact with visual elements.<br/>
    The application must call `resume_control()^` so that users can interact again
    with the visual elements.

    Arguments:
        state (State^): The current user state as received in any callback.
        callback (Optional[Union[str, Callable]]): the function to be called if the user
            chooses to cancel.<br/>
            If empty or None, no cancel action is provided to the user.
        message: the message to show.
    """
    if state and isinstance(state._gui, Gui):
        state._gui._hold_actions(callback, message)
    else:
        warnings.warn("'hold_actions()' must be called in the context of a callback")


def resume_control(state: State):
    """Resume the User Interface actions.

    This function must be called after `hold_control()^` was invoked, when interaction
    must be allowed again for the user.

    Arguments:
        state (State^): The current user state as received in any callback.
    """
    if state and isinstance(state._gui, Gui):
        state._gui._resume_actions()
    else:
        warnings.warn("'resume_actions()' must be called in the context of a callback")


def navigate(state: State, to: t.Optional[str] = ""):
    """Navigate to a page.

    Arguments:
        state (State^): The current user state as received in any callback.
        to: The name of the page to navigate to. This must be a valid page identifier.
            If ommitted, the application navigates to the root page.
    """
    if state and isinstance(state._gui, Gui):
        state._gui._navigate(to)
    else:
        warnings.warn("'navigate()' must be called in the context of a callback")



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import typing as t
import warnings

from ..icon import Icon
from . import _MapDict


class _Adapter:
    def __init__(self):
        self.__adapter_for_type: t.Dict[str, t.Callable] = {}
        self.__type_for_variable: t.Dict[str, str] = {}
        self.__warning_by_type: t.Set[str] = set()

    def _add_for_type(self, type_name: str, adapter: t.Callable) -> None:
        self.__adapter_for_type[type_name] = adapter

    def _add_type_for_var(self, var_name: str, type_name: str) -> None:
        self.__type_for_variable[var_name] = type_name

    def _get_for_type(self, type_name: str) -> t.Optional[t.Callable]:
        return self.__adapter_for_type.get(type_name)

    def _get_unique_type(self, type_name: str) -> str:
        index = 0
        while type_name in self.__adapter_for_type:
            type_name = f"{type_name}{index}"
            index += 1
        return type_name

    def _run_for_var(self, var_name: str, value: t.Any, id_only=False) -> t.Any:
        ret = self._run(self.__get_for_var(var_name, value), value, var_name, id_only)
        if ret is not None:
            return ret
        return value

    def __get_for_var(self, var_name: str, value: t.Any) -> t.Optional[t.Callable]:
        adapter = None
        type_name = self.__type_for_variable.get(var_name)
        if not isinstance(type_name, str):
            adapter = self.__adapter_for_type.get(var_name)
            type_name = var_name if callable(adapter) else type(value).__name__
        if adapter is None:
            adapter = self.__adapter_for_type.get(type_name)
        if callable(adapter):
            return adapter
        return None

    def _get_elt_per_ids(self, var_name: str, lov: t.List[t.Any]) -> t.Dict[str, t.Any]:
        dict_res = {}
        adapter = self.__get_for_var(var_name, lov[0] if lov else None)
        for value in lov:
            try:
                result = adapter(value._dict if isinstance(value, _MapDict) else value) if adapter else value
                dict_res[self.__get_id(result)] = value
                children = self.__get_children(result)
                if children is not None:
                    dict_res.update(self._get_elt_per_ids(var_name, children))
            except Exception as e:
                warnings.warn(f"Can't run adapter for {var_name}: {e}")
        return dict_res

    def _run(
        self, adapter: t.Optional[t.Callable], value: t.Any, var_name: str, id_only=False
    ) -> t.Union[t.Tuple[str, ...], str, None]:
        if value is None:
            return None
        try:
            result = value._dict if isinstance(value, _MapDict) else value
            if adapter:
                result = adapter(result)
            elif isinstance(result, str):
                return result
            result = self._get_valid_result(result, id_only)
            if result is None:
                warnings.warn(
                    f"Adapter for {var_name} did not return a valid result. Please check the documentation on List of Values Adapters."
                )
            else:
                if not id_only and len(result) > 2 and isinstance(result[2], list) and len(result[2]) > 0:
                    result = (result[0], result[1], self.__on_tree(adapter, result[2]))
                return result
        except Exception as e:
            warnings.warn(f"Can't run adapter for {var_name}: {e}")
        return None

    def __on_tree(self, adapter: t.Optional[t.Callable], tree: t.List[t.Any]):
        ret_list = []
        for elt in tree:
            ret = self._run(adapter, elt, adapter.__name__ if adapter else "adapter")
            if ret is not None:
                ret_list.append(ret)
        return ret_list

    def _get_valid_result(self, value: t.Any, id_only=False) -> t.Union[t.Tuple[str, ...], str, None]:
        id = self.__get_id(value)
        if id_only:
            return id
        label = self.__get_label(value)
        if label is None:
            return None
        children = self.__get_children(value)
        return (id, label) if children is None else (id, label, children)  # type: ignore

    def __get_id(self, value: t.Any, dig = True) -> str:
        if isinstance(value, str):
            return value
        elif dig:
            if isinstance(value, (list, tuple)) and len(value):
                return self.__get_id(value[0], False)
            elif hasattr(value, "id"):
                return self.__get_id(value.id, False)
            elif hasattr(value, "__getitem__") and "id" in value:
                return self.__get_id(value.get("id"), False)
        if type(value).__name__ not in self.__warning_by_type:
            warnings.warn(f"LoV id must be a string, using a string representation of {type(value)}")
            self.__warning_by_type.add(type(value).__name__)
        return str(value)

    def __get_label(self, value: t.Any, dig = True) -> t.Union[str, t.Dict, None]:
        if isinstance(value, (str, Icon)):
            return Icon.get_dict_or(value)
        elif dig:
            if isinstance(value, (list, tuple)) and len(value) > 1:
                return self.__get_label(value[1], False)
            elif hasattr(value, "label"):
                return self.__get_label(value.label, False)
            elif hasattr(value, "__getitem__") and "label" in value:
                return self.__get_label(value["label"], False)
        return None

    def __get_children(self, value: t.Any) -> t.Union[t.List[t.Any], None]:
        if isinstance(value, (tuple, list)) and len(value) > 2:
            return value[2] if isinstance(value[2], list) else [value[2]]
        elif hasattr(value, "children"):
            return value.children if isinstance(value.children, list) else [value.children]
        elif hasattr(value, "__getitem__") and "children" in value:
            return value["children"] if isinstance(value["children"], list) else [value["children"]]
        return None



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Any

from markdown.extensions import Extension

from .blocproc import _StartBlockProcessor
from .control import _ControlPattern
from .postproc import _Postprocessor
from .preproc import _Preprocessor


class _TaipyMarkdownExtension(Extension):

    config = {"gui": ["", "Gui object for extension"]}

    def extendMarkdown(self, md):
        from ...gui import Gui

        gui = self.config["gui"][0]
        if not isinstance(gui, Gui):
            raise RuntimeError("Gui instance is not bound to Markdown Extension")
        md.registerExtension(self)
        _Preprocessor.extend(md, gui, 210)
        _ControlPattern.extend(md, gui, 205)
        _StartBlockProcessor.extend(md, gui, 175)
        _Postprocessor.extend(md, gui, 200)



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t


def _is_boolean_true(s: t.Union[bool, str]) -> bool:
    return s if isinstance(s, bool) else s.lower() in ["true", "1", "t", "y", "yes", "yeah", "sure"]


def _is_boolean(s: t.Any) -> bool:
    if isinstance(s, bool):
        return True
    elif isinstance(s, str):
        return s.lower() in ["true", "1", "t", "y", "yes", "yeah", "sure", "false", "0", "f", "no"]
    else:
        return False
